@@ -54,6 +54,7 @@
                                          string-take
                                          string-suffix?
                                          string-index
+                                         string-every
                                          string-join
                                          string-trim-right
                                          string-trim-both
@@ -139,6 +140,10 @@
 ;; This problem occurs when the @code{#\:} (colon), which normally
 ;; separates the header name from its value, is not found.
 ;; The argument is a pair @code{(@var{line} . @var{headers}).
+;;
+;; @item no-name
+;; This problem occurs if there is nothing before the colon except
+;; whitespace.  The argument is as for @code{missing-colon}.
 ;; @end table
 ;;
 (define (read-headers port norm)
@@ -176,6 +181,9 @@
             (else
              (let ((colon (string-index line #\:)))
                (or colon (badness 'missing-colon (cons line (racc!))))
+               (and (or (zero? colon)
+                        (string-every char-set:whitespace line 0 colon))
+                    (badness 'no-name (cons line (racc!))))
                (loop (acons
                       (norm (string-trim-right line char-set:whitespace 0 colon))
                       (string-trim-both line char-set:whitespace (1+ colon))
@@ -200,6 +208,11 @@
       (lambda (string)
         (string->symbol (s2s string)))))
 
+(define (unravel x)                     ; ZONKME 2013-05-15
+  (if (thunk? x)
+      (x)
+      x))
+
 (define (get-body-proc sock hsym headers)
 
   (define (string-read!/partial s)
@@ -232,7 +245,7 @@
   (define (motion options)
     (cond ((memq 'custom options)
            => (lambda (ls)
-                (let-values (((mkx r! cat-r subseq) (cadr ls)))
+                (let-values (((mkx r! cat-r subseq) (unravel (cadr ls))))
                   (values mkx
                           (lambda (len)
                             (let ((x (mkx len)))
@@ -245,10 +258,11 @@
                           cat-r
                           subseq))))
           ((memq 'u8 options)
-           (motion (append `(custom ,(values make-u8vector
-                                             uniform-vector-read!
-                                             u8-concatenate-reverse
-                                             sub-u8))
+           (motion (append `(custom ,(lambda ()
+                                       (values make-u8vector
+                                               uniform-vector-read!
+                                               u8-concatenate-reverse
+                                               sub-u8)))
                            (delq 'u8 options))))
           (else
            (values make-string
@@ -266,44 +280,48 @@
             reverse!
             concat-reverse))
 
-      (define (try-chunked)
+      (define (chunked?)
         (let ((t-enc (assq (hsym "Transfer-Encoding") headers)))
           (and (pair? t-enc)
                (equal? "chunked" (cdr t-enc))
-               (let loop ((acc '()))
-                 (let* ((spec (read-through-CRLF sock))
-                        (len (string->number
-                              (cond ((string-index spec #\;)
-                                     => (lambda (semi)
-                                          ;; ignore extensions for now
-                                          (string-take spec semi)))
-                                    (else spec))
-                              16)))
-                   (if (positive? len)
-                       ;; more
-                       (let* ((chunk (data-in! len))
-                              (end (in! 2)))
-                         (or (string=? CRLF end)
-                             (throw 'chunked-transfer-encoding
-                                    'trailing-garbage end))
-                         (loop (cons chunk acc)))
-                       ;; done
-                       (values
-                        (append
-                         ;; Self-deprecate when work is done.
-                         (delq t-enc headers)
-                         ;; New trailers, if any.
-                         (read-headers sock hsym))
-                        (body<-acc acc))))))))
+               t-enc)))
+
+      (define (handle-chunked t-enc)
+        (let loop ((acc '()))
+          (let* ((spec (read-through-CRLF sock))
+                 (len (string->number
+                       (cond ((string-index spec #\;)
+                              => (lambda (semi)
+                                   ;; ignore extensions for now
+                                   (string-take spec semi)))
+                             (else spec))
+                       16)))
+            (if (positive? len)
+                ;; more
+                (let* ((chunk (data-in! len))
+                       (end (in! 2)))
+                  (or (string=? CRLF end)
+                      (throw 'chunked-transfer-encoding
+                             'trailing-garbage end))
+                  (loop (cons chunk acc)))
+                ;; done
+                (values
+                 (append
+                  ;; Self-deprecate when work is done.
+                  (delq t-enc headers)
+                  ;; New trailers, if any.
+                  (read-headers sock hsym))
+                 (body<-acc acc))))))
 
       (define (same-headers s)
         (values #f s))
 
-      (define (try-known)
-        (and=> (assq-ref headers (hsym "Content-Length"))
-               (lambda (s)
-                 (same-headers
-                  (data-in! (string->number s))))))
+      (define (known?)
+        (assq-ref headers (hsym "Content-Length")))
+
+      (define (handle-known s)
+        (same-headers
+         (data-in! (string->number s))))
 
       (define (drain)
         (let loop ((acc '()))
@@ -318,9 +336,11 @@
                    (same-headers
                     (body<-acc acc)))))))
 
-      (or (try-chunked)
-          (try-known)
-          (drain))))
+      ;; Use ‘cond’ instead of ‘or’ w/ combined test/handling because
+      ;; the rv is two values, which ‘or’ cannot handle under Guile 2.x.
+      (cond ((chunked?) => handle-chunked)
+            ((known?) => handle-known)
+            (else (drain)))))
 
   get-body)
 
